from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from pytest_lazy_fixtures.lazy_fixture import lf as lazy_fixture

from semantic_release.cli.commands.changelog import changelog
from semantic_release.cli.commands.generate_config import generate_config
from semantic_release.cli.commands.main import main
from semantic_release.cli.commands.publish import publish
from semantic_release.cli.commands.version import version

from tests.const import MAIN_PROG_NAME, SUCCESS_EXIT_CODE
from tests.fixtures.repos import repo_w_trunk_only_conventional_commits
from tests.util import assert_exit_code

if TYPE_CHECKING:
    from click import Command

    from tests.conftest import RunCliFn
    from tests.fixtures import UpdatePyprojectTomlFn
    from tests.fixtures.git_repo import BuiltRepoResult


# Define the expected exit code for the help command
HELP_EXIT_CODE = SUCCESS_EXIT_CODE


@pytest.mark.parametrize(
    "help_option", ("-h", "--help"), ids=lambda opt: opt.lstrip("-")
)
@pytest.mark.parametrize(
    "command",
    (main, changelog, generate_config, publish, version),
    ids=lambda cmd: cmd.name,
)
def test_help_no_repo(
    help_option: str,
    command: Command,
    run_cli: RunCliFn,
    change_to_ex_proj_dir: None,
):
    """
    Test that the help message is displayed even when the current directory is not a git repository
    and there is not a configuration file available.
    Documented issue #840
    """
    # Generate some expected output that should be specific per command
    cmd_usage = str.join(
        " ",
        list(
            filter(
                None,
                [
                    "Usage:",
                    MAIN_PROG_NAME,
                    command.name if command.name != "main" else "",
                    "[OPTIONS]",
                    "" if command.name != main.name else "COMMAND [ARGS]...",
                ],
            )
        ),
    )

    # Create the arguments list for subcommands unless its main
    args = list(
        filter(None, [command.name if command.name != main.name else "", help_option])
    )

    # Run the command with the help option
    result = run_cli(args, invoke_kwargs={"prog_name": MAIN_PROG_NAME})

    # Evaluate result
    assert_exit_code(HELP_EXIT_CODE, result, [MAIN_PROG_NAME, *args])
    assert cmd_usage in result.output


@pytest.mark.parametrize(
    "help_option", ("-h", "--help"), ids=lambda opt: opt.lstrip("-")
)
@pytest.mark.parametrize(
    "command",
    (main, changelog, generate_config, publish, version),
    ids=lambda cmd: cmd.name,
)
@pytest.mark.usefixtures(repo_w_trunk_only_conventional_commits.__name__)
def test_help_valid_config(
    help_option: str,
    command: Command,
    run_cli: RunCliFn,
):
    """
    Test that the help message is displayed when the current directory is a git repository
    and there is a valid configuration file available.
    Documented issue #840
    """
    cmd_usage = str.join(
        " ",
        list(
            filter(
                None,
                [
                    "Usage:",
                    MAIN_PROG_NAME,
                    command.name if command.name != main.name else "",
                    "[OPTIONS]",
                    "" if command.name != main.name else "COMMAND [ARGS]...",
                ],
            )
        ),
    )

    # Create the arguments list for subcommands unless its main
    args = list(
        filter(None, [command.name if command.name != main.name else "", help_option])
    )

    # Run the command with the help option
    result = run_cli(args, invoke_kwargs={"prog_name": MAIN_PROG_NAME})

    # Evaluate result
    assert_exit_code(HELP_EXIT_CODE, result, [MAIN_PROG_NAME, *args])
    assert cmd_usage in result.output


@pytest.mark.parametrize(
    "help_option", ("-h", "--help"), ids=lambda opt: opt.lstrip("-")
)
@pytest.mark.parametrize(
    "command",
    (main, changelog, generate_config, publish, version),
    ids=lambda cmd: cmd.name,
)
@pytest.mark.usefixtures(repo_w_trunk_only_conventional_commits.__name__)
def test_help_invalid_config(
    help_option: str,
    command: Command,
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
):
    """
    Test that the help message is displayed when the current directory is a git repository
    and there is an invalid configuration file available.
    Documented issue #840
    """
    # Update the configuration file to have an invalid value
    update_pyproject_toml("tool.semantic_release.remote.type", "invalidhvcs")

    # Generate some expected output that should be specific per command
    cmd_usage = str.join(
        " ",
        list(
            filter(
                None,
                [
                    "Usage:",
                    MAIN_PROG_NAME,
                    command.name if command.name != "main" else "",
                    "[OPTIONS]",
                    "" if command.name != main.name else "COMMAND [ARGS]...",
                ],
            )
        ),
    )

    # Create the arguments list for subcommands unless its main
    args = list(
        filter(None, [command.name if command.name != main.name else "", help_option])
    )

    # Run the command with the help option
    result = run_cli(args, invoke_kwargs={"prog_name": MAIN_PROG_NAME})

    # Evaluate result
    assert_exit_code(HELP_EXIT_CODE, result, [MAIN_PROG_NAME, *args])
    assert cmd_usage in result.output


@pytest.mark.parametrize(
    "help_option", ("-h", "--help"), ids=lambda opt: opt.lstrip("-")
)
@pytest.mark.parametrize(
    "command",
    (main, changelog, generate_config, publish, version),
    ids=lambda cmd: cmd.name,
)
@pytest.mark.parametrize(
    "repo_result", [lazy_fixture(repo_w_trunk_only_conventional_commits.__name__)]
)
def test_help_non_release_branch(
    help_option: str,
    command: Command,
    run_cli: RunCliFn,
    repo_result: BuiltRepoResult,
):
    """
    Test that the help message is displayed even when the current branch is not a release branch.
    Documented issue #840
    """
    # Create & checkout a non-release branch
    non_release_branch = repo_result["repo"].create_head("feature-branch")
    non_release_branch.checkout()

    # Generate some expected output that should be specific per command
    cmd_usage = str.join(
        " ",
        list(
            filter(
                None,
                [
                    "Usage:",
                    MAIN_PROG_NAME,
                    command.name if command.name != "main" else "",
                    "[OPTIONS]",
                    "" if command.name != main.name else "COMMAND [ARGS]...",
                ],
            )
        ),
    )

    # Create the arguments list for subcommands unless its main
    args = list(
        filter(None, [command.name if command.name != main.name else "", help_option])
    )

    # Run the command with the help option
    result = run_cli(args, invoke_kwargs={"prog_name": MAIN_PROG_NAME})

    # Evaluate result
    assert_exit_code(HELP_EXIT_CODE, result, [MAIN_PROG_NAME, *args])
    assert cmd_usage in result.output



from __future__ import annotations

from typing import TYPE_CHECKING, cast

import pytest
from pytest_lazy_fixtures.lazy_fixture import lf as lazy_fixture

from semantic_release.hvcs.github import Github

from tests.const import (
    MAIN_PROG_NAME,
    VERSION_SUBCMD,
)
from tests.fixtures.commit_parsers import (
    conventional_minor_commits,
    default_conventional_parser,
)
from tests.fixtures.git_repo import get_commit_def_of_conventional_commit
from tests.fixtures.repos import (
    repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits_using_tag_format,
    repo_w_no_tags_conventional_commits,
    repo_w_trunk_only_conventional_commits,
    repo_w_trunk_only_conventional_commits_using_tag_format,
)
from tests.fixtures.repos.trunk_based_dev.repo_w_no_tags import (
    repo_w_no_tags_conventional_commits_using_tag_format,
    repo_w_no_tags_conventional_commits_w_zero_version,
)
from tests.util import (
    add_text_to_file,
    assert_exit_code,
    assert_successful_exit_code,
)

if TYPE_CHECKING:
    from unittest.mock import MagicMock

    from requests_mock import Mocker

    from semantic_release.commit_parser._base import CommitParser, ParserOptions
    from semantic_release.commit_parser.token import ParseResult

    from tests.conftest import RunCliFn
    from tests.e2e.conftest import StripLoggingMessagesFn
    from tests.fixtures.git_repo import (
        BuiltRepoResult,
        GetCfgValueFromDefFn,
        GetCommitDefFn,
        GetVersionsFromRepoBuildDefFn,
        SimulateChangeCommitsNReturnChangelogEntryFn,
    )


@pytest.mark.parametrize(
    "repo_result, commits, force_args, next_release_version",
    [
        (
            lazy_fixture(repo_w_trunk_only_conventional_commits.__name__),
            lazy_fixture(conventional_minor_commits.__name__),
            cli_args,
            next_release_version,
        )
        for cli_args, next_release_version in (
            # Dynamic version bump determination (based on commits)
            ([], "0.2.0"),
            # Dynamic version bump determination (based on commits) with build metadata
            (["--build-metadata", "build.12345"], "0.2.0+build.12345"),
            # Forced version bump
            (["--prerelease"], "0.1.1-rc.1"),
            (["--patch"], "0.1.2"),
            (["--minor"], "0.2.0"),
            (["--major"], "1.0.0"),
            # Forced version bump with --build-metadata
            (["--patch", "--build-metadata", "build.12345"], "0.1.2+build.12345"),
            # Forced version bump with --as-prerelease
            (["--prerelease", "--as-prerelease"], "0.1.1-rc.1"),
            (["--patch", "--as-prerelease"], "0.1.2-rc.1"),
            (["--minor", "--as-prerelease"], "0.2.0-rc.1"),
            (["--major", "--as-prerelease"], "1.0.0-rc.1"),
            # Forced version bump with --as-prerelease and modified --prerelease-token
            (
                ["--patch", "--as-prerelease", "--prerelease-token", "beta"],
                "0.1.2-beta.1",
            ),
            # Forced version bump with --as-prerelease and modified --prerelease-token
            # and --build-metadata
            (
                [
                    "--patch",
                    "--as-prerelease",
                    "--prerelease-token",
                    "beta",
                    "--build-metadata",
                    "build.12345",
                ],
                "0.1.2-beta.1+build.12345",
            ),
        )
    ],
)
def test_version_print_next_version(
    repo_result: BuiltRepoResult,
    commits: list[str],
    force_args: list[str],
    next_release_version: str,
    file_in_repo: str,
    run_cli: RunCliFn,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
):
    """
    Given a generic repository at the latest release version and a subsequent commit,
    When running the version command with the --print flag,
    Then the expected next version should be printed and exit without
    making any changes to the repository.

    Note: The point of this test is to only verify that the `--print` flag does not
    make any changes to the repository--not to validate if the next version is calculated
    correctly per the repository structure (see test_version_release &
    test_version_force_level for correctness).

    However, we do validate that --print & a force option and/or --as-prerelease options
    work together to print the next version correctly but not make a change to the repo.
    """
    repo = repo_result["repo"]

    # Make a commit to ensure we have something to release
    # otherwise the "no release will be made" logic will kick in first
    add_text_to_file(repo, file_in_repo)
    repo.git.commit(m=commits[-1], a=True)

    # Setup: take measurement before running the version command
    repo_status_before = repo.git.status(short=True)
    head_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--print", *force_args]
    result = run_cli(cli_cmd[1:], env={Github.DEFAULT_ENV_TOKEN_NAME: "1234"})

    # take measurement after running the version command
    repo_status_after = repo.git.status(short=True)
    head_after = repo.head.commit.hexsha
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)
    assert f"{next_release_version}\n" == result.stdout

    # assert nothing else happened (no code changes, no commit, no tag, no push, no vcs release)
    assert repo_status_before == repo_status_after
    assert head_before == head_after
    assert not tags_set_difference
    assert mocked_git_push.call_count == 0
    assert post_mocker.call_count == 0


@pytest.mark.parametrize(
    "repo_result, commits, force_args, next_release_version",
    [
        *[
            pytest.param(
                lazy_fixture(repo_fixture_name),
                lazy_fixture(conventional_minor_commits.__name__),
                cli_args,
                next_release_version,
                marks=marks if marks else [],
            )
            for repo_fixture_name, marks in (
                (repo_w_trunk_only_conventional_commits.__name__, None),
                (
                    repo_w_trunk_only_conventional_commits_using_tag_format.__name__,
                    pytest.mark.comprehensive,
                ),
            )
            for cli_args, next_release_version in (
                # Dynamic version bump determination (based on commits)
                ([], "0.2.0"),
                # Dynamic version bump determination (based on commits) with build metadata
                (["--build-metadata", "build.12345"], "0.2.0+build.12345"),
                # Forced version bump
                (["--prerelease"], "0.1.1-rc.1"),
                (["--patch"], "0.1.2"),
                (["--minor"], "0.2.0"),
                (["--major"], "1.0.0"),
                # Forced version bump with --build-metadata
                (["--patch", "--build-metadata", "build.12345"], "0.1.2+build.12345"),
                # Forced version bump with --as-prerelease
                (["--prerelease", "--as-prerelease"], "0.1.1-rc.1"),
                (["--patch", "--as-prerelease"], "0.1.2-rc.1"),
                (["--minor", "--as-prerelease"], "0.2.0-rc.1"),
                (["--major", "--as-prerelease"], "1.0.0-rc.1"),
                # Forced version bump with --as-prerelease and modified --prerelease-token
                (
                    ["--patch", "--as-prerelease", "--prerelease-token", "beta"],
                    "0.1.2-beta.1",
                ),
                # Forced version bump with --as-prerelease and modified --prerelease-token
                # and --build-metadata
                (
                    [
                        "--patch",
                        "--as-prerelease",
                        "--prerelease-token",
                        "beta",
                        "--build-metadata",
                        "build.12345",
                    ],
                    "0.1.2-beta.1+build.12345",
                ),
            )
        ],
        *[
            pytest.param(
                lazy_fixture(repo_fixture_name),
                [],
                cli_args,
                next_release_version,
                marks=pytest.mark.comprehensive,
            )
            for repo_fixture_name in (
                repo_w_no_tags_conventional_commits_w_zero_version.__name__,
            )
            for cli_args, next_release_version in (
                # Dynamic version bump determination (based on commits)
                ([], "0.1.0"),
                # Dynamic version bump determination (based on commits) with build metadata
                (["--build-metadata", "build.12345"], "0.1.0+build.12345"),
                # Forced version bump
                (["--prerelease"], "0.0.0-rc.1"),
                (["--patch"], "0.0.1"),
                (["--minor"], "0.1.0"),
                (["--major"], "1.0.0"),
                # Forced version bump with --build-metadata
                (["--patch", "--build-metadata", "build.12345"], "0.0.1+build.12345"),
                # Forced version bump with --as-prerelease
                (["--prerelease", "--as-prerelease"], "0.0.0-rc.1"),
                (["--patch", "--as-prerelease"], "0.0.1-rc.1"),
                (["--minor", "--as-prerelease"], "0.1.0-rc.1"),
                (["--major", "--as-prerelease"], "1.0.0-rc.1"),
                # Forced version bump with --as-prerelease and modified --prerelease-token
                (
                    ["--patch", "--as-prerelease", "--prerelease-token", "beta"],
                    "0.0.1-beta.1",
                ),
                # Forced version bump with --as-prerelease and modified --prerelease-token
                # and --build-metadata
                (
                    [
                        "--patch",
                        "--as-prerelease",
                        "--prerelease-token",
                        "beta",
                        "--build-metadata",
                        "build.12345",
                    ],
                    "0.0.1-beta.1+build.12345",
                ),
            )
        ],
    ],
)
def test_version_print_tag_prints_next_tag(
    repo_result: BuiltRepoResult,
    commits: list[str],
    force_args: list[str],
    next_release_version: str,
    get_cfg_value_from_def: GetCfgValueFromDefFn,
    file_in_repo: str,
    run_cli: RunCliFn,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
):
    """
    Given a generic repository at the latest release version and a subsequent commit,
    When running the version command with the --print-tag flag,
    Then the expected next release tag should be printed and exit without
    making any changes to the repository.

    Note: The point of this test is to only verify that the `--print-tag` flag does not
    make any changes to the repository--not to validate if the next version is calculated
    correctly per the repository structure (see test_version_release &
    test_version_force_level for correctness).

    However, we do validate that --print-tag & a force option and/or --as-prerelease options
    work together to print the next release tag correctly but not make a change to the repo.
    """
    repo = repo_result["repo"]
    repo_def = repo_result["definition"]
    tag_format_str: str = get_cfg_value_from_def(repo_def, "tag_format_str")  # type: ignore[assignment]
    next_release_tag = tag_format_str.format(version=next_release_version)

    if len(commits) > 1:
        # Make a commit to ensure we have something to release
        # otherwise the "no release will be made" logic will kick in first
        add_text_to_file(repo, file_in_repo)
        repo.git.commit(m=commits[-1], a=True)

    # Setup: take measurement before running the version command
    repo_status_before = repo.git.status(short=True)
    head_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--print-tag", *force_args]
    result = run_cli(cli_cmd[1:], env={Github.DEFAULT_ENV_TOKEN_NAME: "1234"})

    # take measurement after running the version command
    repo_status_after = repo.git.status(short=True)
    head_after = repo.head.commit.hexsha
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)
    assert f"{next_release_tag}\n" == result.stdout

    # assert nothing else happened (no code changes, no commit, no tag, no push, no vcs release)
    assert repo_status_before == repo_status_after
    assert head_before == head_after
    assert not tags_set_difference
    assert mocked_git_push.call_count == 0
    assert post_mocker.call_count == 0


@pytest.mark.parametrize(
    "repo_result, commits, force_args, next_release_version",
    [
        pytest.param(
            lazy_fixture(repo_fixture_name),
            [],
            cli_args,
            next_release_version,
            marks=pytest.mark.comprehensive,
        )
        for repo_fixture_name in (
            repo_w_no_tags_conventional_commits.__name__,
            repo_w_no_tags_conventional_commits_using_tag_format.__name__,
        )
        for cli_args, next_release_version in (
            # Dynamic version bump determination (based on commits)
            ([], "1.0.0"),
            # Dynamic version bump determination (based on commits) with build metadata
            (["--build-metadata", "build.12345"], "1.0.0+build.12345"),
            # Forced version bump
            (["--prerelease"], "0.0.0-rc.1"),
            (["--patch"], "0.0.1"),
            (["--minor"], "0.1.0"),
            (["--major"], "1.0.0"),
            # Forced version bump with --build-metadata
            (["--patch", "--build-metadata", "build.12345"], "0.0.1+build.12345"),
            # Forced version bump with --as-prerelease
            (["--prerelease", "--as-prerelease"], "0.0.0-rc.1"),
            (["--patch", "--as-prerelease"], "0.0.1-rc.1"),
            (["--minor", "--as-prerelease"], "0.1.0-rc.1"),
            (["--major", "--as-prerelease"], "1.0.0-rc.1"),
            # Forced version bump with --as-prerelease and modified --prerelease-token
            (
                ["--patch", "--as-prerelease", "--prerelease-token", "beta"],
                "0.0.1-beta.1",
            ),
            # Forced version bump with --as-prerelease and modified --prerelease-token
            # and --build-metadata
            (
                [
                    "--patch",
                    "--as-prerelease",
                    "--prerelease-token",
                    "beta",
                    "--build-metadata",
                    "build.12345",
                ],
                "0.0.1-beta.1+build.12345",
            ),
        )
    ],
)
def test_version_print_tag_prints_next_tag_no_zero_versions(
    repo_result: BuiltRepoResult,
    commits: list[str],
    force_args: list[str],
    next_release_version: str,
    get_cfg_value_from_def: GetCfgValueFromDefFn,
    file_in_repo: str,
    run_cli: RunCliFn,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
):
    """
    Given a generic repository at the latest release version and a subsequent commit,
    When running the version command with the --print-tag flag,
    Then the expected next release tag should be printed and exit without
    making any changes to the repository.

    Note: The point of this test is to only verify that the `--print-tag` flag does not
    make any changes to the repository--not to validate if the next version is calculated
    correctly per the repository structure (see test_version_release &
    test_version_force_level for correctness).

    However, we do validate that --print-tag & a force option and/or --as-prerelease options
    work together to print the next release tag correctly but not make a change to the repo.
    """
    repo = repo_result["repo"]
    repo_def = repo_result["definition"]
    tag_format_str: str = get_cfg_value_from_def(repo_def, "tag_format_str")  # type: ignore[assignment]
    next_release_tag = tag_format_str.format(version=next_release_version)

    if len(commits) > 1:
        # Make a commit to ensure we have something to release
        # otherwise the "no release will be made" logic will kick in first
        add_text_to_file(repo, file_in_repo)
        repo.git.commit(m=commits[-1], a=True)

    # Setup: take measurement before running the version command
    repo_status_before = repo.git.status(short=True)
    head_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--print-tag", *force_args]
    result = run_cli(cli_cmd[1:], env={Github.DEFAULT_ENV_TOKEN_NAME: "1234"})

    # take measurement after running the version command
    repo_status_after = repo.git.status(short=True)
    head_after = repo.head.commit.hexsha
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)
    assert f"{next_release_tag}\n" == result.stdout

    # assert nothing else happened (no code changes, no commit, no tag, no push, no vcs release)
    assert repo_status_before == repo_status_after
    assert head_before == head_after
    assert not tags_set_difference
    assert mocked_git_push.call_count == 0
    assert post_mocker.call_count == 0


@pytest.mark.parametrize(
    "repo_result",
    [lazy_fixture(repo_w_trunk_only_conventional_commits.__name__)],
)
def test_version_print_last_released_prints_version(
    repo_result: BuiltRepoResult,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    run_cli: RunCliFn,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    strip_logging_messages: StripLoggingMessagesFn,
):
    repo = repo_result["repo"]
    latest_release_version = get_versions_from_repo_build_def(
        repo_result["definition"]
    )[-1]

    # Setup: take measurement before running the version command
    repo_status_before = repo.git.status(short=True)
    head_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--print-last-released"]
    result = run_cli(cli_cmd[1:], env={Github.DEFAULT_ENV_TOKEN_NAME: "1234"})

    # take measurement after running the version command
    repo_status_after = repo.git.status(short=True)
    head_after = repo.head.commit.hexsha
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)
    assert not strip_logging_messages(result.stderr)
    assert f"{latest_release_version}\n" == result.stdout

    # assert nothing else happened (no code changes, no commit, no tag, no push, no vcs release)
    assert repo_status_before == repo_status_after
    assert head_before == head_after
    assert not tags_set_difference
    assert mocked_git_push.call_count == 0
    assert post_mocker.call_count == 0


@pytest.mark.parametrize(
    "repo_result, commits",
    [
        (
            lazy_fixture(repo_w_trunk_only_conventional_commits.__name__),
            lazy_fixture(conventional_minor_commits.__name__),
        )
    ],
)
def test_version_print_last_released_prints_released_if_commits(
    repo_result: BuiltRepoResult,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    commits: list[str],
    run_cli: RunCliFn,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    file_in_repo: str,
    strip_logging_messages: StripLoggingMessagesFn,
):
    repo = repo_result["repo"]
    latest_release_version = get_versions_from_repo_build_def(
        repo_result["definition"]
    )[-1]

    # Make a commit so the head is not on the last release
    add_text_to_file(repo, file_in_repo)
    repo.git.commit(m=commits[0], a=True)

    # Setup: take measurement before running the version command
    repo_status_before = repo.git.status(short=True)
    head_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--print-last-released"]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    repo_status_after = repo.git.status(short=True)
    head_after = repo.head.commit.hexsha
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)
    assert not strip_logging_messages(result.stderr)
    assert f"{latest_release_version}\n" == result.stdout

    # assert nothing else happened (no code changes, no commit, no tag, no push, no vcs release)
    assert repo_status_before == repo_status_after
    assert head_before == head_after
    assert not tags_set_difference
    assert mocked_git_push.call_count == 0
    assert post_mocker.call_count == 0


@pytest.mark.parametrize(
    "repo_result",
    [lazy_fixture(repo_w_no_tags_conventional_commits.__name__)],
)
def test_version_print_last_released_prints_nothing_if_no_tags(
    repo_result: BuiltRepoResult,
    run_cli: RunCliFn,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
):
    repo = repo_result["repo"]

    # Setup: take measurement before running the version command
    repo_status_before = repo.git.status(short=True)
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--print-last-released"]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    repo_status_after = repo.git.status(short=True)
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate (no release actions should have occurred on print)
    assert_successful_exit_code(result, cli_cmd)
    assert result.stdout == ""
    assert "No release tags found." in result.stderr

    # assert nothing else happened (no code changes, no commit, no tag, no push, no vcs release)
    assert repo_status_before == repo_status_after
    assert head_sha_before == head_after.hexsha  # No commit has been made
    assert not tags_set_difference  # No tag created
    assert mocked_git_push.call_count == 0  # no git push of tag or commit
    assert post_mocker.call_count == 0  # no vcs release


@pytest.mark.parametrize(
    "repo_result",
    [lazy_fixture(repo_w_trunk_only_conventional_commits.__name__)],
)
def test_version_print_last_released_on_detached_head(
    repo_result: BuiltRepoResult,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    run_cli: RunCliFn,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    strip_logging_messages: StripLoggingMessagesFn,
):
    repo = repo_result["repo"]
    latest_release_version = get_versions_from_repo_build_def(
        repo_result["definition"]
    )[-1]

    # Setup: put the repo in a detached head state
    repo.git.checkout("HEAD", detach=True)

    # Setup: take measurement before running the version command
    repo_status_before = repo.git.status(short=True)
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--print-last-released"]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    repo_status_after = repo.git.status(short=True)
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate (expected -> actual)
    assert_successful_exit_code(result, cli_cmd)
    assert not strip_logging_messages(result.stderr)
    assert f"{latest_release_version}\n" == result.stdout

    # assert nothing else happened (no code changes, no commit, no tag, no push, no vcs release)
    assert repo_status_before == repo_status_after
    assert head_sha_before == head_after.hexsha  # No commit has been made
    assert not tags_set_difference  # No tag created
    assert mocked_git_push.call_count == 0  # no git push of tag or commit
    assert post_mocker.call_count == 0  # no vcs release


@pytest.mark.parametrize(
    "repo_result",
    [lazy_fixture(repo_w_trunk_only_conventional_commits.__name__)],
)
def test_version_print_last_released_on_nonrelease_branch(
    repo_result: BuiltRepoResult,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    run_cli: RunCliFn,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    strip_logging_messages: StripLoggingMessagesFn,
):
    repo = repo_result["repo"]
    latest_release_version = get_versions_from_repo_build_def(
        repo_result["definition"]
    )[-1]

    # Setup: put the repo on a non-release branch
    repo.create_head("next").checkout()

    # Setup: take measurement before running the version command
    repo_status_before = repo.git.status(short=True)
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--print-last-released"]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    repo_status_after = repo.git.status(short=True)
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate (expected -> actual)
    assert_successful_exit_code(result, cli_cmd)
    assert not strip_logging_messages(result.stderr)
    assert f"{latest_release_version}\n" == result.stdout

    # assert nothing else happened (no code changes, no commit, no tag, no push, no vcs release)
    assert repo_status_before == repo_status_after
    assert head_sha_before == head_after.hexsha  # No commit has been made
    assert not tags_set_difference  # No tag created
    assert mocked_git_push.call_count == 0  # no git push of tag or commit
    assert post_mocker.call_count == 0  # no vcs release


@pytest.mark.parametrize(
    "repo_result",
    [
        lazy_fixture(repo_w_trunk_only_conventional_commits.__name__),
        pytest.param(
            lazy_fixture(
                repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits_using_tag_format.__name__
            ),
            marks=pytest.mark.comprehensive,
        ),
    ],
)
def test_version_print_last_released_tag_prints_correct_tag(
    repo_result: BuiltRepoResult,
    get_cfg_value_from_def: GetCfgValueFromDefFn,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    run_cli: RunCliFn,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    strip_logging_messages: StripLoggingMessagesFn,
):
    repo = repo_result["repo"]
    repo_def = repo_result["definition"]
    tag_format_str: str = get_cfg_value_from_def(repo_def, "tag_format_str")  # type: ignore[assignment]
    latest_release_version = get_versions_from_repo_build_def(repo_def)[-1]
    latest_release_tag = tag_format_str.format(version=latest_release_version)

    # Setup: take measurement before running the version command
    repo_status_before = repo.git.status(short=True)
    head_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--print-last-released-tag"]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    repo_status_after = repo.git.status(short=True)
    head_after = repo.head.commit.hexsha
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)
    assert not strip_logging_messages(result.stderr)
    assert f"{latest_release_tag}\n" == result.stdout

    # assert nothing else happened (no code changes, no commit, no tag, no push, no vcs release)
    assert repo_status_before == repo_status_after
    assert head_before == head_after
    assert not tags_set_difference
    assert mocked_git_push.call_count == 0
    assert post_mocker.call_count == 0


@pytest.mark.parametrize(
    "repo_result, commits",
    [
        (
            lazy_fixture(repo_w_trunk_only_conventional_commits.__name__),
            lazy_fixture(conventional_minor_commits.__name__),
        ),
        pytest.param(
            lazy_fixture(
                repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits_using_tag_format.__name__
            ),
            lazy_fixture(conventional_minor_commits.__name__),
            marks=pytest.mark.comprehensive,
        ),
    ],
)
def test_version_print_last_released_tag_prints_released_if_commits(
    repo_result: BuiltRepoResult,
    get_cfg_value_from_def: GetCfgValueFromDefFn,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    commits: list[str],
    run_cli: RunCliFn,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    file_in_repo: str,
    strip_logging_messages: StripLoggingMessagesFn,
):
    repo = repo_result["repo"]
    repo_def = repo_result["definition"]
    tag_format_str: str = get_cfg_value_from_def(repo_def, "tag_format_str")  # type: ignore[assignment]
    latest_release_version = get_versions_from_repo_build_def(repo_def)[-1]
    latest_release_tag = tag_format_str.format(version=latest_release_version)

    # Make a commit so the head is not on the last release
    add_text_to_file(repo, file_in_repo)
    repo.git.commit(m=commits[0], a=True)

    # Setup: take measurement before running the version command
    repo_status_before = repo.git.status(short=True)
    head_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--print-last-released-tag"]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    repo_status_after = repo.git.status(short=True)
    head_after = repo.head.commit.hexsha
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)
    assert not strip_logging_messages(result.stderr)
    assert f"{latest_release_tag}\n" == result.stdout

    # assert nothing else happened (no code changes, no commit, no tag, no push, no vcs release)
    assert repo_status_before == repo_status_after
    assert head_before == head_after
    assert not tags_set_difference
    assert mocked_git_push.call_count == 0
    assert post_mocker.call_count == 0


@pytest.mark.parametrize(
    "repo_result",
    [lazy_fixture(repo_w_no_tags_conventional_commits.__name__)],
)
def test_version_print_last_released_tag_prints_nothing_if_no_tags(
    repo_result: BuiltRepoResult,
    run_cli: RunCliFn,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
):
    repo = repo_result["repo"]

    # Setup: take measurement before running the version command
    repo_status_before = repo.git.status(short=True)
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--print-last-released-tag"]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    repo_status_after = repo.git.status(short=True)
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate (no release actions should have occurred on print)
    assert_successful_exit_code(result, cli_cmd)
    assert result.stdout == ""
    assert "No release tags found." in result.stderr

    # assert nothing else happened (no code changes, no commit, no tag, no push, no vcs release)
    assert repo_status_before == repo_status_after
    assert head_sha_before == head_after.hexsha  # No commit has been made
    assert not tags_set_difference  # No tag created
    assert mocked_git_push.call_count == 0  # no git push of tag or commit
    assert post_mocker.call_count == 0  # no vcs release


@pytest.mark.parametrize(
    "repo_result",
    [
        lazy_fixture(repo_w_trunk_only_conventional_commits.__name__),
        pytest.param(
            lazy_fixture(
                repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits_using_tag_format.__name__
            ),
            marks=pytest.mark.comprehensive,
        ),
    ],
)
def test_version_print_last_released_tag_on_detached_head(
    repo_result: BuiltRepoResult,
    get_cfg_value_from_def: GetCfgValueFromDefFn,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    run_cli: RunCliFn,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    strip_logging_messages: StripLoggingMessagesFn,
):
    repo = repo_result["repo"]
    repo_def = repo_result["definition"]
    tag_format_str: str = get_cfg_value_from_def(repo_def, "tag_format_str")  # type: ignore[assignment]
    latest_release_version = get_versions_from_repo_build_def(repo_def)[-1]
    latest_release_tag = tag_format_str.format(version=latest_release_version)

    # Setup: put the repo in a detached head state
    repo.git.checkout("HEAD", detach=True)

    # Setup: take measurement before running the version command
    repo_status_before = repo.git.status(short=True)
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--print-last-released-tag"]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    repo_status_after = repo.git.status(short=True)
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate (expected -> actual)
    assert_successful_exit_code(result, cli_cmd)
    assert not strip_logging_messages(result.stderr)
    assert f"{latest_release_tag}\n" == result.stdout

    # assert nothing else happened (no code changes, no commit, no tag, no push, no vcs release)
    assert repo_status_before == repo_status_after
    assert head_sha_before == head_after.hexsha  # No commit has been made
    assert not tags_set_difference  # No tag created
    assert mocked_git_push.call_count == 0  # no git push of tag or commit
    assert post_mocker.call_count == 0  # no vcs release


@pytest.mark.parametrize(
    "repo_result",
    [
        lazy_fixture(repo_w_trunk_only_conventional_commits.__name__),
        pytest.param(
            lazy_fixture(
                repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits_using_tag_format.__name__
            ),
            marks=pytest.mark.comprehensive,
        ),
    ],
)
def test_version_print_last_released_tag_on_nonrelease_branch(
    repo_result: BuiltRepoResult,
    get_cfg_value_from_def: GetCfgValueFromDefFn,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    run_cli: RunCliFn,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    strip_logging_messages: StripLoggingMessagesFn,
):
    repo = repo_result["repo"]
    repo_def = repo_result["definition"]
    tag_format_str: str = get_cfg_value_from_def(repo_def, "tag_format_str")  # type: ignore[assignment]
    latest_release_version = get_versions_from_repo_build_def(repo_def)[-1]
    last_release_tag = tag_format_str.format(version=latest_release_version)

    # Setup: put the repo on a non-release branch
    repo.create_head("next").checkout()

    # Setup: take measurement before running the version command
    repo_status_before = repo.git.status(short=True)
    head_sha_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--print-last-released-tag"]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    repo_status_after = repo.git.status(short=True)
    head_after = repo.head.commit
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate (expected -> actual)
    assert_successful_exit_code(result, cli_cmd)
    assert not strip_logging_messages(result.stderr)
    assert f"{last_release_tag}\n" == result.stdout

    # assert nothing else happened (no code changes, no commit, no tag, no push, no vcs release)
    assert repo_status_before == repo_status_after
    assert head_sha_before == head_after.hexsha
    assert not tags_set_difference
    assert mocked_git_push.call_count == 0
    assert post_mocker.call_count == 0


@pytest.mark.parametrize(
    "repo_result, get_commit_def_fn, default_parser",
    [
        (
            lazy_fixture(repo_w_trunk_only_conventional_commits.__name__),
            lazy_fixture(get_commit_def_of_conventional_commit.__name__),
            lazy_fixture(default_conventional_parser.__name__),
        )
    ],
)
def test_version_print_next_version_fails_on_detached_head(
    repo_result: BuiltRepoResult,
    run_cli: RunCliFn,
    simulate_change_commits_n_rtn_changelog_entry: SimulateChangeCommitsNReturnChangelogEntryFn,
    get_commit_def_fn: GetCommitDefFn[CommitParser[ParseResult, ParserOptions]],
    default_parser: CommitParser[ParseResult, ParserOptions],
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    strip_logging_messages: StripLoggingMessagesFn,
):
    repo = repo_result["repo"]
    expected_error_msg = (
        "Detached HEAD state cannot match any release groups; no release will be made"
    )

    # Setup: put the repo in a detached head state
    repo.git.checkout("HEAD", detach=True)

    # Setup: make a commit to ensure we have something to release
    simulate_change_commits_n_rtn_changelog_entry(
        repo,
        [get_commit_def_fn("fix: make a patch fix to codebase", parser=default_parser)],
    )

    # Setup: take measurement before running the version command
    repo_status_before = repo.git.status(short=True)
    head_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--print"]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    repo_status_after = repo.git.status(short=True)
    head_after = repo.head.commit.hexsha
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate (expected -> actual)
    assert_exit_code(1, result, cli_cmd)
    assert not result.stdout
    assert f"{expected_error_msg}\n" == strip_logging_messages(result.stderr)

    # assert nothing else happened (no code changes, no commit, no tag, no push, no vcs release)
    assert repo_status_before == repo_status_after
    assert head_before == head_after
    assert not tags_set_difference
    assert mocked_git_push.call_count == 0
    assert post_mocker.call_count == 0


@pytest.mark.parametrize(
    "repo_result, get_commit_def_fn, default_parser",
    [
        (
            lazy_fixture(repo_w_trunk_only_conventional_commits.__name__),
            lazy_fixture(get_commit_def_of_conventional_commit.__name__),
            lazy_fixture(default_conventional_parser.__name__),
        )
    ],
)
def test_version_print_next_tag_fails_on_detached_head(
    repo_result: BuiltRepoResult,
    run_cli: RunCliFn,
    simulate_change_commits_n_rtn_changelog_entry: SimulateChangeCommitsNReturnChangelogEntryFn,
    get_commit_def_fn: GetCommitDefFn[CommitParser[ParseResult, ParserOptions]],
    default_parser: CommitParser[ParseResult, ParserOptions],
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    strip_logging_messages: StripLoggingMessagesFn,
):
    repo = repo_result["repo"]
    expected_error_msg = (
        "Detached HEAD state cannot match any release groups; no release will be made"
    )

    # Setup: put the repo in a detached head state
    repo.git.checkout("HEAD", detach=True)

    # Setup: make a commit to ensure we have something to release
    simulate_change_commits_n_rtn_changelog_entry(
        repo,
        [get_commit_def_fn("fix: make a patch fix to codebase", parser=default_parser)],
    )

    # Setup: take measurement before running the version command
    repo_status_before = repo.git.status(short=True)
    head_before = repo.head.commit.hexsha
    tags_before = {tag.name for tag in repo.tags}

    # Act
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--print-tag"]
    result = run_cli(cli_cmd[1:])

    # take measurement after running the version command
    repo_status_after = repo.git.status(short=True)
    head_after = repo.head.commit.hexsha
    tags_after = {tag.name for tag in repo.tags}
    tags_set_difference = cast("set[str]", set.difference(tags_after, tags_before))

    # Evaluate (expected -> actual)
    assert_exit_code(1, result, cli_cmd)
    assert not result.stdout
    assert f"{expected_error_msg}\n" == strip_logging_messages(result.stderr)

    # assert nothing else happened (no code changes, no commit, no tag, no push, no vcs release)
    assert repo_status_before == repo_status_after
    assert head_before == head_after
    assert not tags_set_difference
    assert mocked_git_push.call_count == 0
    assert post_mocker.call_count == 0



from __future__ import annotations

import os
import shutil
import sys
from pathlib import Path, PurePosixPath
from re import compile as regexp
from typing import TYPE_CHECKING
from unittest import mock

import pytest
import tomlkit
from pydantic import RootModel, ValidationError
from urllib3.util.url import parse_url

import semantic_release
from semantic_release.cli.config import (
    BranchConfig,
    ChangelogConfig,
    ChangelogOutputFormat,
    GlobalCommandLineOptions,
    HvcsClient,
    RawConfig,
    RuntimeContext,
    _known_hvcs,
)
from semantic_release.cli.util import load_raw_config_file
from semantic_release.commit_parser.conventional import ConventionalCommitParserOptions
from semantic_release.commit_parser.emoji import EmojiParserOptions
from semantic_release.commit_parser.scipy import ScipyParserOptions
from semantic_release.commit_parser.tag import TagParserOptions
from semantic_release.const import DEFAULT_COMMIT_AUTHOR
from semantic_release.enums import LevelBump
from semantic_release.errors import ParserLoadError

from tests.fixtures.repos import repo_w_no_tags_conventional_commits
from tests.util import (
    CustomParserOpts,
    CustomParserWithNoOpts,
    CustomParserWithOpts,
    IncompleteCustomParser,
)

if TYPE_CHECKING:
    from typing import Any

    from tests.fixtures.example_project import ExProjectDir, UpdatePyprojectTomlFn
    from tests.fixtures.git_repo import BuildRepoFn, BuiltRepoResult, CommitConvention


@pytest.mark.parametrize(
    "patched_os_environ, remote_config, expected_token",
    [
        (
            {"GH_TOKEN": "mytoken"},
            {"type": HvcsClient.GITHUB.value},
            "mytoken",
        ),
        (
            {"GITLAB_TOKEN": "mytoken"},
            {"type": HvcsClient.GITLAB.value},
            "mytoken",
        ),
        (
            {"GITEA_TOKEN": "mytoken"},
            {"type": HvcsClient.GITEA.value},
            "mytoken",
        ),
        (
            # default not provided -> means Github
            {"GH_TOKEN": "mytoken"},
            {},
            "mytoken",
        ),
        (
            {"CUSTOM_TOKEN": "mytoken"},
            {"type": HvcsClient.GITHUB.value, "token": {"env": "CUSTOM_TOKEN"}},
            "mytoken",
        ),
    ],
)
def test_load_hvcs_default_token(
    patched_os_environ: dict[str, str],
    remote_config: dict[str, Any],
    expected_token: str,
):
    with mock.patch.dict(os.environ, patched_os_environ, clear=True):
        raw_config = RawConfig.model_validate(
            {
                "remote": remote_config,
            }
        )

    assert expected_token == raw_config.remote.token


@pytest.mark.parametrize("remote_config", [{"type": "nonexistent"}])
def test_invalid_hvcs_type(remote_config: dict[str, Any]):
    with pytest.raises(ValidationError) as excinfo:
        RawConfig.model_validate(
            {
                "remote": remote_config,
            }
        )
    assert "remote.type" in str(excinfo.value)


@pytest.mark.parametrize(
    "commit_parser, expected_parser_opts",
    [
        (
            None,
            RootModel(ConventionalCommitParserOptions()).model_dump(),
        ),  # default not provided -> means conventional
        ("conventional", RootModel(ConventionalCommitParserOptions()).model_dump()),
        ("emoji", RootModel(EmojiParserOptions()).model_dump()),
        ("scipy", RootModel(ScipyParserOptions()).model_dump()),
        ("tag", RootModel(TagParserOptions()).model_dump()),
        (f"{CustomParserWithNoOpts.__module__}:{CustomParserWithNoOpts.__name__}", {}),
        (
            f"{CustomParserWithOpts.__module__}:{CustomParserWithOpts.__name__}",
            RootModel(CustomParserOpts()).model_dump(),
        ),
    ],
)
def test_load_default_parser_opts(
    commit_parser: str | None, expected_parser_opts: dict[str, Any]
):
    raw_config = RawConfig.model_validate(
        # Since TOML does not support NoneTypes, we need to not include the key
        {"commit_parser": commit_parser} if commit_parser else {}
    )
    assert expected_parser_opts == raw_config.commit_parser_options


def test_load_user_defined_parser_opts():
    user_defined_opts = {
        "allowed_tags": ["foo", "bar", "baz"],
        "minor_tags": ["bar"],
        "patch_tags": ["baz"],
        "default_bump_level": LevelBump.PATCH.value,
    }
    raw_config = RawConfig.model_validate(
        {
            "commit_parser": "conventional",
            "commit_parser_options": user_defined_opts,
        }
    )
    assert user_defined_opts == raw_config.commit_parser_options


@pytest.mark.parametrize("commit_parser", [""])
def test_invalid_commit_parser_value(commit_parser: str):
    with pytest.raises(ValidationError) as excinfo:
        RawConfig.model_validate(
            {
                "commit_parser": commit_parser,
            }
        )
    assert "commit_parser" in str(excinfo.value)


def test_default_toml_config_valid(example_project_dir: ExProjectDir):
    default_config_file = example_project_dir / "default.toml"

    default_config_file.write_text(
        tomlkit.dumps(RawConfig().model_dump(mode="json", exclude_none=True))
    )

    written = default_config_file.read_text(encoding="utf-8")
    loaded = tomlkit.loads(written).unwrap()
    # Check that we can load it correctly
    parsed = RawConfig.model_validate(loaded)
    assert parsed
    # Check the re-loaded internal representation is sufficient
    # There is an issue with BaseModel.__eq__ that means
    # comparing directly doesn't work with parsed.dict(); this
    # is because of how tomlkit parsed toml


@pytest.mark.parametrize(
    "mock_env, expected_author",
    [
        ({}, DEFAULT_COMMIT_AUTHOR),
        ({"GIT_COMMIT_AUTHOR": "foo <foo>"}, "foo <foo>"),
    ],
)
@pytest.mark.usefixtures(repo_w_no_tags_conventional_commits.__name__)
def test_commit_author_configurable(
    example_pyproject_toml: Path,
    mock_env: dict[str, str],
    expected_author: str,
    change_to_ex_proj_dir: None,
):
    content = tomlkit.loads(example_pyproject_toml.read_text(encoding="utf-8")).unwrap()

    with mock.patch.dict(os.environ, mock_env):
        raw = RawConfig.model_validate(content)
        runtime = RuntimeContext.from_raw_config(
            raw=raw,
            global_cli_options=GlobalCommandLineOptions(),
        )
        resulting_author = (
            f"{runtime.commit_author.name} <{runtime.commit_author.email}>"
        )
        assert expected_author == resulting_author


def test_load_valid_runtime_config(
    build_configured_base_repo: BuildRepoFn,
    example_project_dir: ExProjectDir,
    example_pyproject_toml: Path,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    change_to_ex_proj_dir: None,
):
    build_configured_base_repo(example_project_dir)

    # Wipe out any existing configuration options
    update_pyproject_toml(f"tool.{semantic_release.__name__}", {})

    runtime_ctx = RuntimeContext.from_raw_config(
        RawConfig.model_validate(load_raw_config_file(example_pyproject_toml)),
        global_cli_options=GlobalCommandLineOptions(),
    )

    # TODO: add more validation
    assert runtime_ctx


@pytest.mark.parametrize(
    "commit_parser",
    [
        # Module:Class string
        f"{CustomParserWithNoOpts.__module__}:{CustomParserWithNoOpts.__name__}",
        f"{CustomParserWithOpts.__module__}:{CustomParserWithOpts.__name__}",
        # File path module:Class string
        f"{CustomParserWithNoOpts.__module__.replace('.', '/')}.py:{CustomParserWithNoOpts.__name__}",
        f"{CustomParserWithOpts.__module__.replace('.', '/')}.py:{CustomParserWithOpts.__name__}",
    ],
)
def test_load_valid_runtime_config_w_custom_parser(
    commit_parser: CommitConvention,
    build_configured_base_repo: BuildRepoFn,
    example_project_dir: ExProjectDir,
    example_pyproject_toml: Path,
    change_to_ex_proj_dir: None,
    request: pytest.FixtureRequest,
):
    fake_sys_modules = {**sys.modules}

    if ".py" in commit_parser:
        module_filepath = Path(commit_parser.split(":")[0])
        module_filepath.parent.mkdir(parents=True, exist_ok=True)
        module_filepath.parent.joinpath("__init__.py").touch()
        shutil.copy(
            src=str(request.config.rootpath / module_filepath),
            dst=str(module_filepath),
        )
        fake_sys_modules.pop(
            str(Path(module_filepath).with_suffix("")).replace(os.sep, ".")
        )

    build_configured_base_repo(
        example_project_dir,
        commit_type=commit_parser,
    )

    with mock.patch.dict(sys.modules, fake_sys_modules, clear=True):
        assert RuntimeContext.from_raw_config(
            RawConfig.model_validate(load_raw_config_file(example_pyproject_toml)),
            global_cli_options=GlobalCommandLineOptions(),
        )


@pytest.mark.parametrize(
    "commit_parser",
    [
        # Non-existant module
        "tests.missing_module:CustomParser",
        # Non-existant class
        f"{CustomParserWithOpts.__module__}:MissingCustomParser",
        # Incomplete class implementation
        f"{IncompleteCustomParser.__module__}:{IncompleteCustomParser.__name__}",
        # Non-existant module file
        "tests/missing_module.py:CustomParser",
        # Non-existant class in module file
        f"{CustomParserWithOpts.__module__.replace('.', '/')}.py:MissingCustomParser",
        # Incomplete class implementation in module file
        f"{IncompleteCustomParser.__module__.replace('.', '/')}.py:{IncompleteCustomParser.__name__}",
    ],
)
def test_load_invalid_custom_parser(
    commit_parser: str,
    build_configured_base_repo: BuildRepoFn,
    example_project_dir: ExProjectDir,
    example_pyproject_toml: Path,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    pyproject_toml_config_option_parser: str,
    change_to_ex_proj_dir: None,
):
    build_configured_base_repo(example_project_dir)

    # Wipe out any existing configuration options
    update_pyproject_toml(f"{pyproject_toml_config_option_parser}_options", {})

    # Insert invalid custom parser string into configuration
    update_pyproject_toml(pyproject_toml_config_option_parser, commit_parser)

    with pytest.raises(ParserLoadError):
        RuntimeContext.from_raw_config(
            RawConfig.model_validate(load_raw_config_file(example_pyproject_toml)),
            global_cli_options=GlobalCommandLineOptions(),
        )


def test_branch_config_with_plain_wildcard():
    branch_config = BranchConfig(
        match="*",
    )
    assert branch_config.match == ".*"


@pytest.mark.parametrize(
    "invalid_regex",
    [
        "*abc",
        "[a-z",
        "(.+",
        "{2,3}",
        "a{3,2}",
    ],
)
def test_branch_config_with_invalid_regex(invalid_regex: str):
    with pytest.raises(ValidationError):
        BranchConfig(
            match=invalid_regex,
        )


@pytest.mark.parametrize(
    "valid_patterns",
    [
        # Single entry
        [r"chore(?:\([^)]*?\))?: .+"],
        # Multiple entries
        [r"^\d+\.\d+\.\d+", r"Initial [Cc]ommit.*"],
    ],
)
def test_changelog_config_with_valid_exclude_commit_patterns(valid_patterns: list[str]):
    assert ChangelogConfig.model_validate(
        {
            "exclude_commit_patterns": valid_patterns,
        }
    )


@pytest.mark.parametrize(
    "invalid_patterns, index_of_invalid_pattern",
    [
        # Single entry, single incorrect
        (["*abc"], 0),
        # Two entries, second incorrect
        ([".*", "[a-z"], 1),
        # Two entries, first incorrect
        (["(.+", ".*"], 0),
    ],
)
def test_changelog_config_with_invalid_exclude_commit_patterns(
    invalid_patterns: list[str],
    index_of_invalid_pattern: int,
):
    with pytest.raises(
        ValidationError,
        match=regexp(
            str.join(
                "",
                [
                    r".*\bexclude_commit_patterns\[",
                    str(index_of_invalid_pattern),
                    r"\]: Invalid regular expression",
                ],
            ),
        ),
    ):
        ChangelogConfig.model_validate(
            {
                "exclude_commit_patterns": invalid_patterns,
            }
        )


@pytest.mark.parametrize(
    "output_format, insertion_flag",
    [
        (
            ChangelogOutputFormat.MARKDOWN.value,
            "<!-- version list -->",
        ),
        (
            ChangelogOutputFormat.RESTRUCTURED_TEXT.value,
            f"..{os.linesep}    version list",
        ),
    ],
)
def test_changelog_config_default_insertion_flag(
    output_format: str,
    insertion_flag: str,
):
    changelog_config = ChangelogConfig.model_validate(
        {
            "default_templates": {
                "output_format": output_format,
            }
        }
    )

    assert changelog_config.insertion_flag == insertion_flag


@pytest.mark.parametrize(
    "hvcs_type",
    [k.value for k in _known_hvcs],
)
def test_git_remote_url_w_insteadof_alias(
    repo_w_initial_commit: BuiltRepoResult,
    example_pyproject_toml: Path,
    example_git_https_url: str,
    hvcs_type: str,
    update_pyproject_toml: UpdatePyprojectTomlFn,
):
    expected_url = parse_url(example_git_https_url)
    repo_name_suffix = PurePosixPath(expected_url.path or "").name
    insteadof_alias = "psr_test_insteadof"
    insteadof_value = expected_url.url.replace(repo_name_suffix, "")
    repo = repo_w_initial_commit["repo"]

    with repo.config_writer() as cfg:
        # Setup: define the insteadOf replacement value
        cfg.add_value(f'url "{insteadof_value}"', "insteadof", f"{insteadof_alias}:")

        # Setup: set the remote URL with an insteadOf alias
        cfg.set_value('remote "origin"', "url", f"{insteadof_alias}:{repo_name_suffix}")

    # Setup: set each supported HVCS client type
    update_pyproject_toml("tool.semantic_release.remote.type", hvcs_type)

    # Act: load the configuration (in clear environment)
    with mock.patch.dict(os.environ, {}, clear=True):
        # Essentially the same as CliContextObj._init_runtime_ctx()
        project_config = tomlkit.loads(
            example_pyproject_toml.read_text(encoding="utf-8")
        ).unwrap()

        runtime = RuntimeContext.from_raw_config(
            raw=RawConfig.model_validate(
                project_config.get("tool", {}).get("semantic_release", {}),
            ),
            global_cli_options=GlobalCommandLineOptions(),
        )

        # Trigger a function that calls helpers.parse_git_url()
        actual_url = runtime.hvcs_client.remote_url(use_token=False)

    # Evaluate: the remote URL should be the full URL
    assert expected_url.url == actual_url



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import pytest
from pytest_lazy_fixtures.lazy_fixture import lf as lazy_fixture

from semantic_release.changelog.context import ChangelogMode

from tests.const import CHANGELOG_SUBCMD, MAIN_PROG_NAME
from tests.fixtures.repos import repo_w_no_tags_conventional_commits
from tests.util import (
    CustomConventionalParserWithIgnorePatterns,
    assert_successful_exit_code,
)

if TYPE_CHECKING:
    from pathlib import Path

    from semantic_release.commit_parser.conventional.parser import (
        ConventionalCommitParser,
    )

    from tests.conftest import RunCliFn
    from tests.fixtures.example_project import UpdatePyprojectTomlFn, UseCustomParserFn
    from tests.fixtures.git_repo import BuiltRepoResult, GetCommitDefFn


@pytest.mark.parametrize(
    "repo_result", [lazy_fixture(repo_w_no_tags_conventional_commits.__name__)]
)
def test_changelog_custom_parser_remove_from_changelog(
    repo_result: BuiltRepoResult,
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    use_custom_parser: UseCustomParserFn,
    get_commit_def_of_conventional_commit: GetCommitDefFn[ConventionalCommitParser],
    changelog_md_file: Path,
    default_md_changelog_insertion_flag: str,
    default_conventional_parser: ConventionalCommitParser,
):
    """
    Given when a changelog filtering custom parser is configured
    When provided a commit message that matches the ignore syntax
    Then the commit message is not included in the resulting changelog
    """
    ignored_commit_def = get_commit_def_of_conventional_commit(
        "chore: do not include me in the changelog",
        parser=default_conventional_parser,
    )

    # Because we are in init mode, the insertion flag is not present in the changelog
    # we must take it out manually because our repo generation fixture includes it automatically
    with changelog_md_file.open(newline=os.linesep) as rfd:
        # use os.linesep here because the insertion flag is os-specific
        # but convert the content to universal newlines for comparison
        expected_changelog_content = (
            rfd.read()
            .replace(f"{default_md_changelog_insertion_flag}{os.linesep}", "")
            .replace("\r", "")
        )

    # Set the project configurations
    update_pyproject_toml(
        "tool.semantic_release.changelog.mode", ChangelogMode.INIT.value
    )
    use_custom_parser(
        f"{CustomConventionalParserWithIgnorePatterns.__module__}:{CustomConventionalParserWithIgnorePatterns.__name__}"
    )

    # Setup: add the commit to be ignored
    repo_result["repo"].git.commit(m=ignored_commit_def["msg"], a=True)

    # Act
    cli_cmd = [MAIN_PROG_NAME, CHANGELOG_SUBCMD]
    result = run_cli(cli_cmd[1:])

    # Take measurement after action
    actual_content = changelog_md_file.read_text()

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)

    # Verify that the changelog content does not include our commit
    assert ignored_commit_def["desc"] not in actual_content

    # Verify that the changelog content has not changed
    assert expected_changelog_content == actual_content



from __future__ import annotations

from pathlib import Path
from shutil import rmtree
from typing import TYPE_CHECKING

import pytest
from git import Repo

import tests.conftest
import tests.const
import tests.fixtures.git_repo
import tests.util
from tests.const import (
    DEFAULT_BRANCH_NAME,
    EXAMPLE_HVCS_DOMAIN,
    EXAMPLE_PROJECT_NAME,
)
from tests.util import copy_dir_tree

if TYPE_CHECKING:
    from typing import Protocol, Sequence

    from git import Actor

    from semantic_release.hvcs import HvcsBase

    from tests.conftest import (
        BuildRepoOrCopyCacheFn,
        GetMd5ForSetOfFilesFn,
        RepoActions,
    )
    from tests.fixtures.git_repo import (
        BuildRepoFn,
        CommitConvention,
        TomlSerializableTypes,
    )

    class BuildMonorepoFn(Protocol):
        def __call__(self, dest_dir: Path | str) -> Path: ...


@pytest.fixture(scope="session")
def deps_files_4_example_git_monorepo(
    deps_files_4_example_monorepo: list[Path],
) -> list[Path]:
    return [
        *deps_files_4_example_monorepo,
        # This file
        Path(__file__).absolute(),
        # because of imports
        Path(tests.const.__file__).absolute(),
        Path(tests.util.__file__).absolute(),
        # because of the fixtures
        Path(tests.conftest.__file__).absolute(),
        Path(tests.fixtures.git_repo.__file__).absolute(),
    ]


@pytest.fixture(scope="session")
def build_spec_hash_4_example_git_monorepo(
    get_md5_for_set_of_files: GetMd5ForSetOfFilesFn,
    deps_files_4_example_git_monorepo: list[Path],
) -> str:
    # Generates a hash of the build spec to set when to invalidate the cache
    return get_md5_for_set_of_files(deps_files_4_example_git_monorepo)


@pytest.fixture(scope="session")
def cached_example_git_monorepo(
    build_repo_or_copy_cache: BuildRepoOrCopyCacheFn,
    build_spec_hash_4_example_git_monorepo: str,
    cached_example_monorepo: Path,
    example_git_https_url: str,
    commit_author: Actor,
) -> Path:
    """
    Initializes an example monorepo project with git. DO NOT USE DIRECTLY.

    Use a `repo_*` fixture instead. This creates a default
    base repository, all settings can be changed later through from the
    example_project_git_repo fixture's return object and manual adjustment.
    """

    def _build_repo(cached_repo_path: Path) -> Sequence[RepoActions]:
        if not cached_example_monorepo.exists():
            raise RuntimeError("Unable to find cached monorepo files")

        # make a copy of the example monorepo as a base
        copy_dir_tree(cached_example_monorepo, cached_repo_path)

        # initialize git repo (open and close)
        # NOTE: We don't want to hold the repo object open for the entire test session,
        # the implementation on Windows holds some file descriptors open until close is called.
        with Repo.init(cached_repo_path) as repo:
            rmtree(str(Path(repo.git_dir, "hooks")))

            # set up remote origin (including a refs directory)
            git_origin_dir = Path(repo.common_dir, "refs", "remotes", "origin")
            repo.create_remote(name=git_origin_dir.name, url=example_git_https_url)
            git_origin_dir.mkdir(parents=True, exist_ok=True)

            # Without this the global config may set it to "master", we want consistency
            repo.git.branch("-M", DEFAULT_BRANCH_NAME)

            with repo.config_writer("repository") as config:
                config.set_value("user", "name", commit_author.name)
                config.set_value("user", "email", commit_author.email)
                config.set_value("commit", "gpgsign", False)
                config.set_value("tag", "gpgsign", False)

                # set up a remote tracking branch for the default branch
                config.set_value(f'branch "{DEFAULT_BRANCH_NAME}"', "remote", "origin")
                config.set_value(
                    f'branch "{DEFAULT_BRANCH_NAME}"',
                    "merge",
                    f"refs/heads/{DEFAULT_BRANCH_NAME}",
                )

            # make sure all base files are in index to enable initial commit
            repo.index.add(("*", ".gitignore"))

        # This is a special build, we don't expose the Repo Actions to the caller
        return []

    # End of _build_repo()

    return build_repo_or_copy_cache(
        repo_name=cached_example_git_monorepo.__name__.split("_", maxsplit=1)[1],
        build_spec_hash=build_spec_hash_4_example_git_monorepo,
        build_repo_func=_build_repo,
    )


@pytest.fixture(scope="session")
def file_in_pkg_pattern(file_in_repo: str, monorepo_pkg_dir_pattern: str) -> str:
    return str(Path(monorepo_pkg_dir_pattern) / file_in_repo)


@pytest.fixture(scope="session")
def file_in_monorepo_pkg1(
    monorepo_pkg1_name: str,
    file_in_pkg_pattern: str,
) -> Path:
    return Path(file_in_pkg_pattern.format(pkg_name=monorepo_pkg1_name))


@pytest.fixture(scope="session")
def file_in_monorepo_pkg2(
    monorepo_pkg2_name: str,
    file_in_pkg_pattern: str,
) -> Path:
    return Path(file_in_pkg_pattern.format(pkg_name=monorepo_pkg2_name))


@pytest.fixture(scope="session")
def build_base_monorepo(  # noqa: C901
    cached_example_git_monorepo: Path,
) -> BuildMonorepoFn:
    """
    This fixture is intended to simplify repo scenario building by initially
    creating the repo but also configuring semantic_release in the pyproject.toml
    for when the test executes semantic_release. It returns a function so that
    derivative fixtures can call this fixture with individual parameters.
    """

    def _build_configured_base_monorepo(dest_dir: Path | str) -> Path:
        if not cached_example_git_monorepo.exists():
            raise RuntimeError("Unable to find cached git project files!")

        # Copy the cached git project the dest directory
        copy_dir_tree(cached_example_git_monorepo, dest_dir)

        return Path(dest_dir)

    return _build_configured_base_monorepo


@pytest.fixture(scope="session")
def configure_monorepo_package(  # noqa: C901
    configure_base_repo: BuildRepoFn,
) -> BuildRepoFn:
    """
    This fixture is intended to simplify repo scenario building by initially
    creating the repo but also configuring semantic_release in the pyproject.toml
    for when the test executes semantic_release. It returns a function so that
    derivative fixtures can call this fixture with individual parameters.
    """

    def _configure(  # noqa: C901
        dest_dir: Path | str,
        commit_type: CommitConvention = "conventional",
        hvcs_client_name: str = "github",
        hvcs_domain: str = EXAMPLE_HVCS_DOMAIN,
        tag_format_str: str | None = None,
        extra_configs: dict[str, TomlSerializableTypes] | None = None,
        mask_initial_release: bool = True,  # Default as of v10
        package_name: str = EXAMPLE_PROJECT_NAME,
        monorepo: bool = True,
    ) -> tuple[Path, HvcsBase]:
        if not monorepo:
            raise ValueError("This fixture is only for monorepo packages!")

        if not Path(dest_dir).exists():
            raise RuntimeError(f"Destination directory {dest_dir} does not exist!")

        return configure_base_repo(
            dest_dir=dest_dir,
            commit_type=commit_type,
            hvcs_client_name=hvcs_client_name,
            hvcs_domain=hvcs_domain,
            tag_format_str=tag_format_str,
            extra_configs=extra_configs,
            mask_initial_release=mask_initial_release,
            package_name=package_name,
            monorepo=monorepo,
        )

    return _configure
